//  g++-mp-9  hadamard_commented.cpp  nauty/nauty.a nauty/naugstrings.c -O3    && time ./a.out  
//  g++  hadamard_match.cpp  nauty/nauty.a nauty/naugstrings.c -O3  -march=native  -o b.out && time ./a.out  
// time parallel -a data263.txt --pipepart -d '\n' -j 60  --block 300000  --progress  ./a.out    > /dev/null 
// time parallel -a all_hadamard.txt --pipepart -d '\n' -j 60  --block 300000  --progress  ./a.out    > all_of_them.txt 

// Final line for lomikamen
// time parallel -a all_hadamard.txt --pipepart -d '\n' -j 55  --block 257000  --progress  ./b.out    > all_indexes.txt 
// One line has about size 257, so block 257000 is about 1000 lines
//
//  13710027/1000/60/60/24 * 35
//  where 35 is the time to finish one computation (parallel will tell the estimate prograss)
//  gives expected time - it is about 1 week
//
// For compilation, one needs to include nauty - used for finding a canonical labelling.
//   - download nauty
//   - extract nauty
//   - do ./configure and make in nauty
//   - make ln -s nautyXXXX nauty
//   - compile this program -see on top how to compile it
//
// This program was first written in Sage by Steve Butler and this is a rewrite to C++ 
// by Bernard Lidicky, which gave about 20x speed up and makes the whole problem computable.
// The Sage program is included in comments and for each Sage block, an equivalent in C++ is below.
//
// Usage: 
//    The program can be used as
//       - reading line by line from stdin  (cin) and processing them
//     or
//       - processing each argument as one matrix

#include <iostream>
#include <unordered_set>
#include <unordered_map> 
#include <vector> 
#include <string>
#include <utility>
#include <iomanip> 
#include <algorithm>
#include <cmath>
#include <cassert>
#include <fstream>

//#define MAXN 32 
// recommended by manual to use MAXN WORDSIZE
#define MAXN WORDSIZE 
#include  "nauty/nauty.h"
#include  "nauty/naututil.h"
#include  "nauty/naugstrings.h"

//#define MAXN 32

using namespace std;



vector<string> all_known_graphs;


void load_all_known_graphs(const char *filename)
{
    ifstream in;
    in.open(filename, std::ifstream::in);

    string graph;
    while (std::getline(in, graph))
    {

        if (graph.length() > 0)
        {
            if (graph[graph.length()-1] != '\n')
            {
                graph.append("\n");
            }
            all_known_graphs.push_back(graph);
        }        
    }
    in.close();

    // Test if the graphs are sorted
    if (!std::is_sorted(all_known_graphs.begin(),all_known_graphs.end()))
    {
        cerr << "Graphs on input are not sroted!!" << endl;

        std::sort(all_known_graphs.begin(),all_known_graphs.end());

        for (auto &a : all_known_graphs)
        {
            cout << a << endl;
        }
        assert(0);
    }

    //cerr << "Loaded " << all_known_graphs.size() << " known graphs" << endl;

}

int find_index_of_a_graph(const string &s)
{
    auto it = lower_bound(all_known_graphs.begin(), all_known_graphs.end(), s);
    if (all_known_graphs.end() == it)
    {
        cerr << "Graph " << s << " not found in the list!!!" << endl;
        return -1;
    }
    return distance(all_known_graphs.begin(),it);
}






inline int get_entry(unsigned long long int entry, int id)
{
    unsigned long long int one = 1;
//    return  (((entry & (1 << id)) >> id) << 1) - 1;
//    return  (((entry & (1 << id)) >> id) << 1) - 1;
    if (entry & (one << id))
        return 1;
    return -1;
}

inline void set_bit_entry(unsigned long long int &entry, int id)
{
    const unsigned long long int one = 1;
    entry = entry | (one << id);
//    return  (((entry & (1 << id)) >> id) << 1) - 1;
//    return  (((entry & (1 << id)) >> id) << 1) - 1;
//    if (entry & (one << id))
//        return 1;
//    return -1;
}

inline int get_bit_entry(const unsigned long long int entry, int id)
{
    const unsigned long long int one = 1;
    return (entry & (one << id)) >> id;
}


inline int count_bits(unsigned long long int entry)
{
    const unsigned long long int one = 1;
    unsigned int count = 0; 

    while (entry) 
    { 
        count += entry & one; 
        entry >>= 1; 
    } 
    return count; 
}    
 
// def recurse_find_hadamard_graphs(partial_sum, columns, stop_check, current_column, edges, generated_graphs):
void recurse_find_hadamard_graphs(int *partial_sum, 
    const vector<vector<int> > &columns, 
    vector<unordered_set<int> > *stop_check, 
    int current_column, 
    const vector< vector<pair<int,int> > >  & edges, 
    unordered_set<int> & generated_graphs,
    int all_entries_size,
    int N)
{

    /* 
    for (auto x : edges)
    {
        for (auto y : x)
        {
            cerr << y.first << " " << y.second << ",";
        }
        cerr << endl;
    }
    return;
    */
/*     
    # The heart of the computation, a recursive routine which searchs
    # through all combinations of columns, and for those that that
    # produce a valid graph adds to the list of generated graphs.
*/

/* 
    if current_column==31:
        # We have reached a leaf in our decision tree; time to decide
        # if this node is a graph.
        if set(partial_sum).issubset(set([0,32])):
            # We have produced a graph; now to find which one.
            graph_edges = []
            for i in range(len(partial_sum)):
                if partial_sum[i] == 32:
                    # The entries corresponding to this term are of the
                    # correct form to be an edge, add in all edges
                    # this will create
                    graph_edges.extend(edges[i])

            # Now make the graph
            G = Graph(32) # create empty graph on 32 vertices
            G.add_edges(graph_edges) # add in the edges
            
            # Following gives unique(!) string for the graph
            # which we then add into the set of generated graphs
            generated_graphs.add(G.canonical_label().graph6_string())

*/
    if (current_column == N-1)
    {
        for (int i = 0; i < all_entries_size; i++)
        {
            if (partial_sum[i] != 0 && partial_sum[i] != N) return;
        }



        DEFAULTOPTIONS_GRAPH(options);
        options.getcanon = TRUE;

        graph g[MAXN*MAXM];
        int lab[MAXN],ptn[MAXN],orbits[MAXN];
        statsblk stats;
        graph g_canon[MAXN*MAXM];
        int n,m,v;
        n = N;
        m = SETWORDSNEEDED(n);
        nauty_check(WORDSIZE,m,n,NAUTYVERSIONID);

        EMPTYGRAPH(g,m,n);

        for (int i =0 ; i < all_entries_size; i++)
        {
            if (partial_sum[i] == N)
            {
                for (const auto &edge : edges[i])
                {
                    ADDONEEDGE(g,edge.first,edge.second,m);
                }
            }
        }

        densenauty(g,lab,ptn,orbits,&options,&stats,m,n,g_canon);

        //extern void putcanon(FILE*,int*,graph*,int,int,int);
        //putgraph(f,canong,linelength,m,n);

        char buff[1024];
        char *pp = buff;
        gtog6string(g_canon, &pp, m, n);
        *pp = 0;

        //cerr << buff << endl;

        generated_graphs.insert( find_index_of_a_graph( string(buff) ));

        return;
    }        

/* 
    else:
        # First, we check what happens if we do not add the current column
        for i in range(31,len(partial_sum)):
            if partial_sum[i] not in stop_check[i][current_column]:
                # If we don't add the current column, then it is
                # impossible to get a graph regardless of choices
                # made further down the tree, do not progress
                break
        else:
            # Entrywise it is still possible to end with a graph
            # so keep recursing further down
            recurse_find_hadamard_graphs(partial_sum, columns, stop_check, current_column+1, edges, generated_graphs)
*/
    bool can_be_finished = true;
    for (int i = N-1; i < all_entries_size; i++)
    {
        if (stop_check[i][current_column].find(partial_sum[i]) == stop_check[i][current_column].end())
        {
            can_be_finished = false;
            break;
        }
    }
    if (can_be_finished)
    {
        recurse_find_hadamard_graphs(partial_sum, columns, stop_check, current_column+1, edges, generated_graphs, all_entries_size, N);
    }
/*
        # Second, we check what happens if we do add the current column
        new_partial_sum = [partial_sum[i]+columns[current_column][i] for i in range(len(partial_sum))]
        for i in range(31,len(partial_sum)):
            if partial_sum[i] not in stop_check[i][current_column]:
                # If we do add the current column, then it is
                # impossible to get a graph regardless of choices
                # made further down the tree, do not progress
                break
        else:
            # Entrywise it is still possible to end with a graph
            # so keep recursing further down
            recurse_find_hadamard_graphs(new_partial_sum, columns, stop_check, current_column+1, edges, generated_graphs)
*/
    // modify the partial_sum
    for (int i = 0; i < all_entries_size; i++)
    {
        partial_sum[i] += columns[current_column][i];
    }

    can_be_finished = true;
    for (int i = N-1; i < all_entries_size; i++)
    {
        if (stop_check[i][current_column].find(partial_sum[i]) == stop_check[i][current_column].end())
        {
            can_be_finished = false;
            break;
        }
    }
    if (can_be_finished)
    {
        recurse_find_hadamard_graphs(partial_sum, columns, stop_check, current_column+1, edges, generated_graphs, all_entries_size, N);
    }

    // clean the  the partial_sum modification
    for (int i = 0; i < all_entries_size; i++)
    {
        partial_sum[i] -= columns[current_column][i];
    }


}


void presort(vector<vector<int> > &columns, int N)
{
/* 
def presort(columns):
    # We presort the columns by trying to put lots of problems early
    # Note the first 31 rows are uninteresting, so we look past them
    h = len(columns[0]) # the height of the columns
    sort_order = [] # the final sorting order
*/

    int h = columns[0].size();

    vector<int> sort_order;
    unordered_set<int> in_sort_order;

    while (sort_order.size() < N-1)
    {

        //cerr << sort_order.size() << " " << N-1 << "   ";
        //for (auto i : sort_order) cerr << " " << i;
        //cerr << endl;

        unordered_map<unsigned long long int, int> sort_options;

        unsigned long long int all_unused = 0;
        for (int i = 0; i < N-1; i++)
            if (in_sort_order.find(i) == in_sort_order.end())
                set_bit_entry(all_unused, i);
        
        //cerr << "all_unused " <<  all_unused << endl;

        sort_options.insert(make_pair(all_unused, 1));

        for (int i = 31; i < h; i++)
        {
            unsigned long long int test = 0;
            for (int j = 0; j < 31; j++)
            {
                if (columns[j][i] && in_sort_order.find(j) == in_sort_order.end())
                {
                    set_bit_entry(test, j);
                }
            }

            if (test == 0)
                continue;

            if (sort_options.find(test) != sort_options.end())
            {
                sort_options.find(test)->second++;
            }
            else
            {
                sort_options.insert(make_pair(test,1));
            }
        }
    
/* 

    while len(sort_order)<31: # still need to make a decision on sorting
        # Basic idea == look for fewest number of columns that would
        # create a problem AND among those find ones which effect the
        # most entries
        sort_options = {}
        sort_options[tuple([i for i in range(31) if i not in sort_order])] = 1
        for i in range(31,h):
            # test stores which unsorted columns are still involved
            test = tuple([j for j in range(31) if (columns[j][i] and j not in sort_order)])
            if len(test): # make sure something to be sorted!
                if test in sort_options:
                    sort_options[test] += 1
                else:
                    sort_options[test] = 1
           */

          /* 
        # sort_options gives us possible ways to sort; want to find
        # one which is used most often (= more likely to cause conflict)
        # we now run through sort_options and find a best candidate
        
        few_columns = 32
        most_rows = 0
        best = []
        for S in sort_options:
            if len(S) == few_columns:
                if sort_options[S] > most_rows:
                    most_rows = sort_options[S]
                    best = S
            elif len(S) < few_columns:
                few_columns = len(S)
                most_rows = sort_options[S]
                best = S
                
        # Now extend the sort by ordering found
        sort_order.extend(list(best))
        */
        int few_columns = N;
        int most_rows = 0;
        unsigned long long int best;

        for (const auto &S : sort_options)
        {
            int cb  = count_bits(S.first);
            if (cb < few_columns || (cb == few_columns && S.second > most_rows))
            {
                few_columns = cb;
                most_rows = S.second;
                best = S.first;
            }
        }

        for (int i = 0; i < N; i++)
        {
            if (get_bit_entry(best, i))
            {
                sort_order.push_back(i);
                in_sort_order.insert(i);
            }
        }
    }
    /* 
    #Return the columns resorted according to new order
    return [columns[i] for i in sort_order]
*/
    vector<vector<int> > tmp_columns = std::move(columns);

    columns.clear();
    for (int i : sort_order)
    {
        columns.push_back(tmp_columns[i]);
    }
}


void find_hadamard_graphs(const char* S,  unordered_set<int> &all_generated_graphs, int Sid, int N)
{
    // S = a hexadecimal string of length 256 used to store the matrix
    // First step is to convert the hexadecimal string to a hadamard matrix
    //translate = {'0':[-1,-1,-1,-1], '1':[-1,-1,-1, 1], '2':[-1,-1, 1,-1], '3':[-1,-1, 1, 1], '4':[-1, 1,-1,-1], '5':[-1, 1,-1, 1], '6':[-1, 1, 1,-1], '7':[-1, 1, 1, 1], '8':[ 1,-1,-1,-1], '9':[ 1,-1,-1, 1], 'A':[ 1,-1, 1,-1], 'B':[ 1,-1, 1, 1], 'C':[ 1, 1,-1,-1], 'D':[ 1, 1,-1, 1], 'E':[ 1, 1, 1,-1], 'F':[ 1, 1, 1, 1]}
    //H = []
    //for s in S:
    //    H.extend(translate[s])
    //H = matrix([H[i*32:(i+1)*32] for i in range(32)])
    
    int Hvec[N*N];
    int HvecID = 0;
    while (*S != 0)
    {
        unsigned int Sn = *S-'0';
        if (Sn > 9)
        {
            Sn = *S - 'A'+10;
        }
        //cerr << *S << Sn << endl;
        Hvec[HvecID++] = (((Sn)&(1<<3)) >> 2) - 1;
        Hvec[HvecID++] = (((Sn)&(1<<2)) >> 1) - 1;
        Hvec[HvecID++] = ((Sn)&(1<<1)) - 1 ;
        Hvec[HvecID++] = (((Sn)&1) << 1) - 1;
        S++;
    }

    int H[N][N];
    HvecID = 0;
    for (int x = 0; x < N; x++)
    {
       for (int y = 0; y < N; y++)
       {
            H[x][y] = Hvec[HvecID++];
            //cerr << setw(2) << H[x][y] << " ";
       }
       //cerr << endl;
    }

    //return;


    // We now find all possible form of entries of M = H^* \Lambda H
    // these are the form of +/-1 vectors corresponding to diagonal
    // elements of \Lambda.  We will also keep track of the
    // locations for which the various forms occur.
    // These are both lists.
    /* 
    all_entries = []
    locations = []
    for i in range(0,32):
        for j in range(i+1,32):
            entry = [-1*H[k,j]*H[k,i] for k in range(1,32)]  # M_{i,j}
            if entry in all_entries:
                # Have we seen this entry before?  If we have, add
                # to the list of locations for entry
                locations[all_entries.index(entry)].append([i,j])
            else:
                # Have we not seen this entry before?  If not, add
                # a new entry type and note where the location was
                all_entries.append(entry)
                locations.append([[i,j]])
    */
   vector<unsigned long long int>  all_entries;
   unordered_map<unsigned long long int, unsigned long long int>  all_entries_index;
   vector< vector<pair<int,int> > > locations;
   for (int i = 0; i < N; i++)
   {
        for (int j = i+1; j < N; j++)
        {
            // This is kinda hack so entry fits just into one int, where
            // 1 means 1 and 0 means -1
            unsigned long long int entry = 0;
            unsigned long long int one = 1;
            for (int k = 1; k < N; k++)
            {
//                if (-1*H[k][j]*H[k][i] == 1)
                if (-H[k][j]*H[k][i] == 1)
                    entry = entry | (one << (k-1));
            }

            if (all_entries_index.find(entry) != all_entries_index.end())
            {
                int entryID = all_entries_index.find(entry)->second;
                locations[entryID].push_back(make_pair(i,j));
            }
            else
            {
                all_entries.push_back(entry);
                all_entries_index[entry] = all_entries.size()-1;

                vector< pair<int,int> > vv;
                vv.push_back(make_pair(i,j));
                locations.push_back(vv);
            }            
        }
   }

    /* 
    cerr << all_entries.size() << endl;
    for (int i = 0; i < N-1; i++)
    {
        cerr << setw(2) << get_entry(all_entries[100],i) << " ";
    }
    cerr << endl;
    return;
    */

    /* 
    # Important note:  the entries in locations are going to be the
    # edges of the graph.  So ideally should be stored in some format
    # which is friendly to making graphs.
    
    # Big idea == if we specify the entries in the first row, this
    # determines \Lambda and so determines all remaining entries.
    # So we now want to express all entries in terms of the entries
    # in the first row.  This is what following will be used for!
        
    MM = [[1-H[j][i] for j in range(1,32)] for i in range(1,32)]
    */
    int MM[N][N];
    for (int i = 1; i < N; i++)
    {
       for (int j = 1; j < N; j++)
       {
            MM[i-1][j-1] = 1-H[j][i];
           //cerr << MM[i-1][j-1] << " ";
       }
       //cerr << endl;
    }

    //return;

    /* 
    # The i-th row of the following matrix indicates how to write
    # an entry in H^* \Lambda H in terms of an entry on the first
    # row (up to a scaling factor).

    NN = [[sum([MM[i][j]*all_entries[k][j] for j in range(31)]) for i in range(31)] for k in range(len(all_entries))]
    */

    int NN[all_entries.size()][N];

    for (int k = 0; k < (int)all_entries.size(); k++)
    {
        int rowsum = 0;
        for (int i = 0; i < N-1; i++)
        {
            int sumofMM = 0;
            for (int j = 0; j < N-1; j++)
            {
                    sumofMM += MM[i][j]*get_entry(all_entries[k],j);
                    //sumofMM += get_entry(all_entries[k],j);
                    //sumofMM += MM[i][j];
            }
            NN[k][i] = sumofMM;
            //cerr << setw(3) <<  NN[k][i] << " ";
            rowsum += sumofMM;
        }
        //cerr << endl;
        assert(rowsum == N);
        //cerr << "   Sum = " << rowsum << endl;
    }        

   //return;




    /* 
    # Theoretical side note:  At this point what we are looking for
    # is an assignment of values to the first row so that all of the
    # entries of the matrix will be of a proper form.  So we think
    # of the assignment of the first row as a {0,1}-vector and this
    # corresponds to selecting some columns of NN so that the row
    # sums of those columns will be a {0,32}-vector (32 coming from
    # scaling factor hidden inside computation of MM above).
    
    # To shrink the search space we will look at ways to explore
    # it as a tree where we can prune at some point if we know that
    # any terminal nodes below will not result in a vector of the
    # corret form.  First we look for a way to order the columns
    # to cause problems earlier (= smaller part of tree to search)
    
    columns = presort([[NN[j][i] for j in range(len(NN))] for i in range(31)])
    */

    vector<vector<int> > columns;
    for (int i = 0; i < N-1; i++)
    {
        vector<int> one_column;
        one_column.reserve(all_entries.size());
        for (int j = 0; j < all_entries.size(); j++)
        {
            one_column.push_back(NN[j][i]);
        }
        columns.push_back(std::move(one_column));
    }

    presort(columns, N);

    //cerr << "Presort done" << endl;

    // TODO: Presort!

    //return;
   /* 
    # We now store information about what partial sums are possible
    # in each entry; this can be used if we see a partial sum which
    # cannot extend then we stop computation (prune the tree).
    # The first 31 entries can *never* create problems, so we skip.
    
    stop_check = [0]*31
    for r in range(31,len(all_entries)):
        stop = [set([0,32])]
        for i in range(31):
            stop.append(set([x for x in stop[-1]]+[x-columns[30-i][r] for x in stop[-1]]))
        stop.reverse()
        stop_check.append(stop)

    */

    vector<unordered_set<int> > stop_check[all_entries.size()];
    // todo maybe fill idexes 0..31?
    for (int r = N-1; r < all_entries.size(); r++)
    {
        unordered_set<int> stop_last;
        stop_last.insert(0);
        stop_last.insert(N);

        vector<unordered_set<int> > stop;
        stop.push_back(stop_last);

        for (int i = 0; i < N-1; i++)
        {
            unordered_set<int> stop_next(stop_last);
            for (int x : stop_last)
            {
                stop_next.insert(x-columns[N-2-i][r]);
            }
            stop.push_back(stop_next);
            stop_last = std::move(stop_next);
        }
        std::reverse(stop.begin(),stop.end()); 
        stop_check[r] = std::move(stop);
    }

    /* 
    for (int i = 0; i < all_entries.size(); i++)
    {
        for(auto vs :  stop_check[i])
        {
            for(auto s : vs)
            {
                cerr << s << " ";
            }
            cerr << endl;
        }
        cerr << endl;
    }
    cerr << all_entries.size() << endl;
    return;
    */

    /*     
    # Initialize some parameters in preparation for running through
    # our decision tree.
    generated_graphs = set([])
    column_sum = [0]*len(all_entries)    
    */

    int column_sum[all_entries.size()];
    for (int i = 0; i < all_entries.size(); i++)
    {
        column_sum[i] = 0;
    }

    unordered_set<int> generated_graphs;

    /* 
    # Recursively run our decision tree
    recurse_find_hadamard_graphs(column_sum, columns, stop_check, 0, locations, generated_graphs)
    
    return generated_graphs
    */
    recurse_find_hadamard_graphs(column_sum, columns, stop_check, 0, locations, generated_graphs, all_entries.size(), N);


    for (auto s : generated_graphs)
    {
        all_generated_graphs.insert(s);
    }

    // cerr << setw(4) << Sid <<  " found " << generated_graphs.size() << " total " << all_generated_graphs.size() << endl;


}


void find_hadamard_graphs(const char* S,  unordered_set<int> &all_generated_graphs, int Sid)
{
    int Slen  = strlen(S);
    
    int N = (int)(2*sqrt(Slen));
    
    //cerr << "N  = " << N << endl;
    //cerr << "MAXN = " << MAXN << endl;

    find_hadamard_graphs(S, all_generated_graphs, Sid, N);
}



void print_generated_graphs(const unordered_set<int> &all_generated_graphs)
{
    vector<int> agg(all_generated_graphs.begin(), all_generated_graphs.end());

    sort(agg.begin(), agg.end());

    bool add_space = false;
    for (int i : agg)
    {
        if (add_space) cout << " ";
        add_space = true;

        cout << i;
    }
    cout << endl;
}

int main(int argc, char *argv[])
{
    load_all_known_graphs("all_of_them_unique.sorted.txt");
//    load_all_known_graphs("all_of_them_unique.txt");


    if (argc > 1)
    {
        for (int i = 1; i < argc; i++)
        {
            unordered_set<int> all_generated_graphs;
            find_hadamard_graphs(argv[i], all_generated_graphs, i);
            print_generated_graphs(all_generated_graphs);
        }
    }
    else
    {
        int linecnt = 0;
        std::string line;
        while (std::getline(cin, line))
        {
            unordered_set<int> all_generated_graphs;
            find_hadamard_graphs(line.c_str(), all_generated_graphs, ++linecnt);
            print_generated_graphs(all_generated_graphs);
        }
    }
    

    return 0;


}

